from abc import ABC, abstractmethod
import numpy as np
from typing import Union


class NetworkInitializer:
    """
    network initializers.
    """

    def __init__(self, method: str = "random"):
        self.method = method

    def gen_initial_states(self, shape: Union[tuple, list]) -> np.ndarray:
        """
        Generate initial states for the reservoir.

        Parameters:
        - shape (tuple, list): The shape of array to generate

        Returns:
        - np.ndarray: The initialized array
        """
        # returns an array of shape <shape>
        # creates the entries based on different sampling methods
        # when not setting specific values, the range is normalized to abs(1)

        # Convert shape to tuple if it is a scalar or list
        if isinstance(shape, int):
            shape = (shape,)
        elif isinstance(shape, list):
            shape = tuple(shape)

        if self.method == "random":
            init_states = np.random.random(*shape)
        elif self.method == "random_normal":
            init_states = np.random.randn(*shape)
        elif self.method == "ones":
            init_states = np.ones(*shape)
        elif self.method == "zeros":
            init_states = np.zeros(*shape)
        else:
            raise ValueError(
                f"Sampling method {self.method} is unknown for generating initial reservoir states"
            )

        # normalize to max. absolute value of 1
        if self.method != "zeros":
            init_states = init_states / np.max(np.abs(init_states))

        return init_states



import numpy as np
import random
from abc import ABC, abstractmethod

from sklearn.linear_model import Ridge


class Optimizer(ABC):

    def __init__(self, name: str = ""):
        self.name = name
        pass

    @abstractmethod
    def solve(self, A: np.ndarray, b: np.ndarray) -> np.ndarray:
        # expects input A = [(n_batch*n_time), n_nodes], b = [(n_batch*n_time), n_out]
        # returns W_out = [n_nodes, n_out]
        pass


class RidgeSK(Optimizer):
    # solves a linear regression model using sklearn's Ridge method,
    # see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html

    def __init__(self, name: str = "", alpha=1.0):
        super().__init__(name)
        self.alpha = 1.0

    def solve(self, A: np.ndarray, b: np.ndarray) -> np.ndarray:
        clf = Ridge(self.alpha, fit_intercept=False).fit(A, b)
        W_out = clf.coef_.T
        return W_out

    def set_alpha(self, value: float):
        if value < 0:
            raise ValueError("Alpha must be non-negative.")
        self.alpha = value

    def get_alpha(self) -> float:
        return self.alpha


def assign_optimizer(optimizer: str or Optimizer) -> Optimizer:
    """
    Maps names of optimizers to the correct implementation.

    Parameters
    ----------
    optimizer : str or Optimizer
        The name of the optimizer.

    Returns
    -------
    Optimizer
        An instance of the optimizer class corresponding to the given name.

    Raises
    ------
    ValueError
        If the given optimizer name is not implemented.
    """

    if not isinstance(optimizer, str):
        if not isinstance(optimizer, Optimizer):
            raise TypeError(
                f"Optimizer must be a string or Optimizer object! Given optimizer is of type:{type(optimizer)}"
            )

    # maps names of optimizers to the correct implementation.
    if optimizer == "ridge" or optimizer == "Ridge":
        return RidgeSK()

    if isinstance(optimizer, Optimizer):
        return optimizer

    # TODO: add more solvers (sparsity promoting, ...)
    else:
        raise (
            ValueError(
                f"{optimizer} not implemented! Check optimizers.py and assign_optimizers()"
            )
        )



import cProfile
import pstats
import memory_profiler
import psutil
import GPUtil
import time
from contextlib import contextmanager
from functools import wraps
import numpy as np
import matplotlib.pyplot as plt
from typing import Dict, List, Any

class RCProfiler:
    """
    Profiler for Reservoir Computing implementations to track CPU, memory, GPU usage and pruning metrics.
    """
    def __init__(self, model, log_file="rc_profile.log"):
        self.model = model
        self.log_file = log_file
        self.process = psutil.Process()
        self.profiler = cProfile.Profile()
        self.metrics = {
            'memory': [],
            'cpu': [],
            'gpu': [],
            'timestamps': [],
            'reservoir_states': [],
            # Adding pruning-specific metrics
            'pruning': {
                'nodes_removed': [],
                'performance_changes': [],
                'network_sizes': [],
                'memory_per_node': [],
                'time_per_iteration': [],
                'spectral_radius': []
            }
        }
        self.start_time = time.time()
    
    @contextmanager
    def profile_section(self, section_name):
        """Context manager for profiling specific sections of code."""
        start_time = time.time()
        start_mem = self.process.memory_info().rss / 1024 / 1024  # MB
        
        yield
        
        end_time = time.time()
        end_mem = self.process.memory_info().rss / 1024 / 1024
        
        with open(self.log_file, 'a') as f:
            f.write(f"\n=== {section_name} ===\n")
            f.write(f"Duration: {end_time - start_time:.2f} seconds\n")
            f.write(f"Memory change: {end_mem - start_mem:.2f} MB\n")
    
    def profile_method(self, method_name):
        """Decorator for profiling class methods."""
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                with self.profile_section(method_name):
                    return func(*args, **kwargs)
            return wrapper
        return decorator
    
    def start_monitoring(self, interval=1.0):
        """Start continuous resource monitoring."""
        self.monitoring = True
        while self.monitoring:
            timestamp = time.time()
            
            # Memory
            mem_usage = self.process.memory_info().rss / 1024 / 1024
            self.metrics['memory'].append(mem_usage)
            
            # CPU
            cpu_percent = self.process.cpu_percent()
            self.metrics['cpu'].append(cpu_percent)
            
            # GPU if available
            try:
                gpus = GPUtil.getGPUs()
                if gpus:
                    gpu_usage = sum(gpu.load * 100 for gpu in gpus) / len(gpus)
                    self.metrics['gpu'].append(gpu_usage)
            except:
                self.metrics['gpu'].append(0)
                
            self.metrics['timestamps'].append(timestamp)
            
            time.sleep(interval)
    
    def stop_monitoring(self):
        """Stop continuous resource monitoring."""
        self.monitoring = False
    
    def profile_fit(self, X, y, **kwargs):
        """Profile the fit method of the RC model."""
        self.profiler.enable()
        
        with self.profile_section("Model Fitting"):
            history = self.model.fit(X, y, **kwargs)
            
            # Track reservoir states if available
            if hasattr(history, 'res_states'):
                self.metrics['reservoir_states'] = history['res_states']
        
        self.profiler.disable()
        return history

    def profile_fit_prune(self, X, y, loss_metric='mse', max_perf_drop=0.1):
        """Profile the fit_prune method with detailed metrics."""
        self.profiler.enable()
        start_time = time.time()
        
        with open(self.log_file, 'a') as f:
            f.write("\n=== Starting Pruning Process ===\n")
        
        # Store initial state
        initial_state = {
            'memory': self.process.memory_info().rss / 1024 / 1024,
            'network_size': self.model.reservoir_layer.weights.shape[0]
        }
        
        # Wrap the original fit_prune method to collect metrics
        original_fit_prune = self.model.fit_prune
        
        @wraps(original_fit_prune)
        def instrumented_fit_prune(*args, **kwargs):
            iteration = 0
            last_perf = None
            
            def track_pruning_step(node_idx, current_performance):
                nonlocal iteration, last_perf
                
                current_time = time.time()
                current_memory = self.process.memory_info().rss / 1024 / 1024
                current_size = self.model.reservoir_layer.weights.shape[0]
                
                # Calculate metrics
                perf_change = (current_performance - last_perf) if last_perf is not None else 0
                last_perf = current_performance
                
                # Get spectral radius
                try:
                    spec_rad = np.max(np.abs(np.linalg.eigvals(
                        self.model.reservoir_layer.weights)))
                except:
                    spec_rad = None
                
                # Store metrics
                self.metrics['pruning']['nodes_removed'].append(node_idx)
                self.metrics['pruning']['performance_changes'].append(perf_change)
                self.metrics['pruning']['network_sizes'].append(current_size)
                self.metrics['pruning']['memory_per_node'].append(
                    current_memory - initial_state['memory'])
                self.metrics['pruning']['time_per_iteration'].append(
                    current_time - start_time)
                if spec_rad is not None:
                    self.metrics['pruning']['spectral_radius'].append(spec_rad)
                
                # Log step
                self.log_pruning_step(iteration, node_idx, current_performance, 
                                    current_size, current_memory)
                
                iteration += 1
            
            # Inject tracking into model
            self.model._track_pruning = track_pruning_step
            
            # Run original fit_prune
            history = original_fit_prune(*args, **kwargs)
            
            # Clean up
            delattr(self.model, '_track_pruning')
            
            return history
        
        # Temporarily replace method
        self.model.fit_prune = instrumented_fit_prune
        
        try:
            history = self.model.fit_prune(X, y, loss_metric=loss_metric, 
                                         max_perf_drop=max_perf_drop)
        finally:
            # Restore original method
            self.model.fit_prune = original_fit_prune
        
        self.profiler.disable()
        return history
    
    def profile_predict(self, X, **kwargs):
        """Profile the predict method of the RC model."""
        self.profiler.enable()
        
        with self.profile_section("Model Prediction"):
            predictions = self.model.predict(X, **kwargs)
        
        self.profiler.disable()
        return predictions
    
    def log_pruning_step(self, iteration, node_idx, performance, network_size, memory_usage):
        """Log information about a pruning step."""
        with open(self.log_file, 'a') as f:
            f.write(f"""
Pruning Step {iteration}:
- Node Removed: {node_idx}
- Network Size: {network_size}
- Performance: {performance:.4f}
- Memory Usage: {memory_usage:.2f} MB
""")
    
    def plot_metrics(self, save_path=None):
        """Plot all collected metrics including pruning metrics if available."""
        if self.metrics['pruning']['nodes_removed']:
            # Create a figure with two rows of subplots
            fig, axes = plt.subplots(3, 2, figsize=(15, 12))
            fig.suptitle('RC Performance and Pruning Analysis')
            
            # Original metrics
            timestamps = np.array(self.metrics['timestamps']) - self.metrics['timestamps'][0]
            
            # Memory usage
            ax = axes[0, 0]
            ax.plot(timestamps, self.metrics['memory'])
            ax.set_ylabel('Memory Usage (MB)')
            ax.set_xlabel('Time (s)')
            ax.grid(True)
            
            # Network size evolution
            ax = axes[0, 1]
            ax.plot(self.metrics['pruning']['network_sizes'])
            ax.set_ylabel('Network Size')
            ax.set_xlabel('Pruning Step')
            ax.grid(True)
            
            # CPU usage
            ax = axes[1, 0]
            ax.plot(timestamps, self.metrics['cpu'])
            ax.set_ylabel('CPU Usage (%)')
            ax.set_xlabel('Time (s)')
            ax.grid(True)
            
            # Performance changes
            ax = axes[1, 1]
            ax.plot(self.metrics['pruning']['performance_changes'])
            ax.set_ylabel('Performance Change')
            ax.set_xlabel('Pruning Step')
            ax.grid(True)
            
            # GPU usage
            ax = axes[2, 0]
            ax.plot(timestamps, self.metrics['gpu'])
            ax.set_ylabel('GPU Usage (%)')
            ax.set_xlabel('Time (s)')
            ax.grid(True)
            
            # Spectral radius if available
            ax = axes[2, 1]
            if self.metrics['pruning']['spectral_radius']:
                ax.plot(self.metrics['pruning']['spectral_radius'])
                ax.set_ylabel('Spectral Radius')
                ax.set_xlabel('Pruning Step')
                ax.grid(True)
            
        else:
            # Original plotting code for non-pruning metrics
            fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(12, 10), sharex=True)
            
            timestamps = np.array(self.metrics['timestamps']) - self.metrics['timestamps'][0]
            
            ax1.plot(timestamps, self.metrics['memory'])
            ax1.set_ylabel('Memory Usage (MB)')
            ax1.grid(True)
            
            ax2.plot(timestamps, self.metrics['cpu'])
            ax2.set_ylabel('CPU Usage (%)')
            ax2.grid(True)
            
            ax3.plot(timestamps, self.metrics['gpu'])
            ax3.set_ylabel('GPU Usage (%)')
            ax3.set_xlabel('Time (s)')
            ax3.grid(True)
        
        plt.tight_layout()
        
        if save_path:
            plt.savefig(save_path)
        plt.show()
    
    def get_profile_stats(self):
        """Get detailed profiling statistics."""
        stats = pstats.Stats(self.profiler)
        stats.sort_stats('cumulative')
        
        with open(self.log_file, 'a') as f:
            f.write("\n=== Detailed Profile Stats ===\n")
            stats.stream = f
            stats.print_stats()
        
        return stats




"""
Testing the deletion of nodes in the reservoir computer
"""

from matplotlib import pyplot as plt
from pyreco.utils_data import sequence_to_sequence as seq_2_seq
from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.optimizers import RidgeSK
import numpy as np

"""
use case: we train a RC, and then choose to delete a set of nodes from the reservoir.
"""


# get some data
X_train, X_test, y_train, y_test = seq_2_seq(
    name="sine_pred", n_batch=20, n_states=2, n_time=150
)

input_shape = X_train.shape[1:]
output_shape = y_train.shape[1:]

# build a classical RC
model = RC()
model.add(InputLayer(input_shape=input_shape))
model.add(
    RandomReservoirLayer(
        nodes=200, density=0.1, activation="tanh", leakage_rate=0.1, fraction_input=1.0
    ),
)
model.add(ReadoutLayer(output_shape, fraction_out=0.9))

# Compile the model
optim = RidgeSK(alpha=0.5)
model.compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
)

# Train the model
model.fit(X_train, y_train)


print(f"score: \t\t\t{model.evaluate(x=X_test, y=y_test)[0]:.4f}")


"""
Now choose to delete some nodes from the reservoir
"""

nodes_to_delete = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

# delete nodes
model.remove_reservoir_nodes(nodes_to_delete)
model.fit(X_train, y_train)
print(f"score after node removal: \t{model.evaluate(x=X_test, y=y_test)[0]:.4f}")

print(f"nodes in reservoir: \t{model.reservoir_layer.nodes}")
print(f"density of reservoir: \t{model.reservoir_layer.density:.3f}")


"""
Now let's cut reservoir nodes and see how the performance changes
"""
num_node_prune = 3
scores = []
num_nodes = []
remaining_nodes = model.reservoir_layer.nodes
while (remaining_nodes - num_node_prune) > 0:
    nodes_to_delete = np.random.choice(
        model.reservoir_layer.nodes, num_node_prune, replace=False
    ).tolist()
    print(f"nodes to delete: {nodes_to_delete}")
    print(f"nodes before deletion: {model.readout_layer.readout_nodes}")
    model.remove_reservoir_nodes(nodes_to_delete)
    print(f"number of nodes in the RC: {model.reservoir_layer.nodes}\n")
    model.fit(X_train, y_train)
    scores.append(model.evaluate(x=X_test, y=y_test)[0])
    num_nodes.append(model.reservoir_layer.nodes)
    remaining_nodes = model.reservoir_layer.nodes

plt.figure()
plt.plot(num_nodes, scores, "o-")
plt.xlabel("Number of nodes in reservoir")
plt.ylabel("MSE")
plt.title("Loss vs. number of nodes in reservoir")
plt.show()



"""
We will have an abstract Layer class, from which the following layers inherit:

- InputLayer
- ReservoirLayer
    - RandomReservoir
    - RecurrenceReservoir
    - EvolvedReservoir
- ReadoutLayer

"""

from abc import ABC, abstractmethod
import numpy as np
import networkx as nx

from pyreco.utils_networks import (
    gen_ER_graph,
    compute_density,
    get_num_nodes,
    compute_spec_rad,
    remove_nodes_from_graph,
    set_spec_rad
)


# implements the abstract base class
class Layer(ABC):

    @abstractmethod
    def __init__(self):
        self.weights = None  # every layer will have some weights (trainable or not)
        self.name: str = "layer"
        self._is_compiled: bool = False  # flag to check if the overall model is compiled
        pass

    @abstractmethod
    def update_layer_properties(self):
        pass


class InputLayer(Layer):
    # Shape of the read-in weights is: N x n_states, where N is the number of nodes in the reservoir, and n_states is
    # the state dimension of the input (irrespective if a time series or a vector was put in)
    # the actual read-in layer matrix will be created by model.compile()!

    def __init__(self, input_shape):
        # input shape is (n_timesteps, n_states)
        super().__init__()
        self.shape = input_shape
        self.n_time = input_shape[0]
        self.n_states = input_shape[1]
        self.name = "input_layer"

        # some properties of the readin layer
        self.fraction_nonzero_entries: (
            float  # fraction of nonzero entries in the input layer
        )

    def remove_nodes(self, nodes: list):
        # removes a node from the input layer (i.e. if a reservoir node needs to be dropped)

        if not isinstance(nodes, list):
            raise TypeError("Nodes must be provided as a list of indices.")
        if np.max(nodes) > self.weights.shape[0]:
            raise ValueError(
                "Node index exceeds the number of nodes in the input layer."
            )
        if np.min(nodes) < 0:
            raise ValueError("Node index must be positive.")
        if not all(isinstance(x, int) for x in nodes):
            raise ValueError("All entries in the node list must be integers.")

        # remove nodes from [n_reservoir_nodes, n_states] matrix
        self.weights = np.delete(self.weights, nodes, axis=0)

        # update the properties of the input layer
        self.update_layer_properties()

    def update_layer_properties(self):

        # updates the properties of the input layer
        self.fraction_nonzero_entries = (
            np.count_nonzero(self.weights) / self.weights.size
        )


class ReadoutLayer(Layer):

    def __init__(self, output_shape, fraction_out=1.0):
        # expects output_shape = (n_timesteps, n_states)
        super().__init__()
        self.output_shape: tuple = output_shape
        self.n_time = output_shape[0]
        self.n_states = output_shape[1]

        self.fraction_out: float = (
            fraction_out  # fraction of connections to the reservoir
        )
        self.name = "readout_layer"

        self.readout_nodes = []  # list of nodes that are linked to output

    def remove_nodes(self, nodes: list):
        # remove nodes from the readout layer, when the user wants to delete specific reservoir nodes

        # # delete entries that equal the nodes to be removed
        # idx_del = np.where(np.isin(self.readout_nodes, nodes))[0]
        # self.readout_nodes = np.delete(self.readout_nodes, idx_del)
        # self.update_layer_properties()
        raise NotImplementedError("This method is not yet implemented.")

    def update_layer_properties(self):

        # updates the properties of the readout layer
        self.fraction_out = len(self.readout_nodes) / self.weights.shape[0]
        pass


class FeedbackLayer(Layer):
    # is same as InputLayer, but for the feedback connections
    # Shape of the read-in weights is: N x n_states, where N is the number of nodes in the reservoir, and n_states is
    # the state dimension of the input (irrespective if a time series or a vector was put in)
    # the actual read-in layer matrix will be created by mode.compile()!

    def __init__(self, feedback_shape):
        # input shape is (n_timesteps, n_states)
        super().__init__()
        self.shape = feedback_shape
        self.n_time = feedback_shape[0]
        self.n_states = feedback_shape[1]
        self.name = "feedback_layer"

        # some properties of the feedback layer
        self.fraction_nonzero_entries: (
            float  # fraction of nonzero entries in the feedback layer
        )

    def remove_nodes(self, nodes: list):
        # removes a node from the input layer (i.e. if a reservoir node needs to be dropped)

        if not isinstance(nodes, list):
            raise TypeError("Nodes must be provided as a list of indices.")
        if np.max(nodes) > self.weights.shape[0]:
            raise ValueError(
                "Node index exceeds the number of nodes in the input layer."
            )
        if np.min(nodes) < 0:
            raise ValueError("Node index must be positive.")
        if not all(isinstance(x, int) for x in nodes):
            raise ValueError("All entries in the node list must be integers.")

        # remove nodes from [n_reservoir_nodes, n_states] matrix
        self.weights = np.delete(self.weights, nodes, axis=0)

        # update the properties of the input layer
        self.update_layer_properties()

    def update_layer_properties(self):

        # updates the properties of the input layer
        self.fraction_nonzero_entries = (
            np.count_nonzero(self.weights) / self.weights.size
        )

class ReservoirLayer(Layer):  # subclass for the specific reservoir layers

    def __init__(
        self,
        nodes,
        density,
        activation,
        leakage_rate,
        fraction_input,
        init_res_sampling,
        seed: int = 42,
    ):
        super().__init__()
        self.nodes: int = nodes  # number of reservoir nodes
        self.density: float = density
        self.spec_rad = None
        self.activation = activation
        self.leakage_rate = leakage_rate
        self.name = "reservoir_layer"
        self.fraction_input = fraction_input
        self.weights = None  # the adjacency matrix / reservoir network
        self.input_receiving_nodes = None

        # initial reservoir state (will be set later)
        self.initial_res_states = None
        self.init_res_sampling = init_res_sampling

    def activation_fun(self, x: np.ndarray):
        if self.activation == "sigmoid":
            return 1 / (1 + np.exp(-x))
        elif self.activation == "tanh":
            return (np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-x))
        else:
            raise (ValueError(f"unknown activation function {self.activation}!"))

    def set_weights(self, network: np.ndarray):
        # set reservoir network (adjacency matrix) from outside.
        # Updates all related parameters. 
        # MUST be having the same shape as the existing reservoir network.

        if not isinstance(network, np.ndarray):
            raise TypeError("Network must be a numpy array.")
        
        if network.ndim != 2:
            raise ValueError("Network must be a 2D numpy array.")
        
        if network.shape[0] != network.shape[1]:
            raise ValueError("Network must be square (same number of rows and columns).")
        
        # for now, we only accept matrices that have the same shape as the existing reservoir
        if self._is_compiled and (network.shape != self.weights.shape):
            raise ValueError(f"Network must have the same shape as the existing reservoir ({self.weights.shape}).")

        # overwrite the existing adjacency matrix
        self.weights = network

        # update related reservoir properties (density, spectral radius, etc.)
        self.update_layer_properties()

    def set_initial_state(self, r_init: np.ndarray):
        # assigns an initial state to each of the reservoir nodes

        if r_init.shape[0] != self.nodes:
            raise (
                ValueError(
                    "initial reservoir state does not match the number of nodes in the reservoir!"
                )
            )
        self.initial_res_states = r_init

    def remove_nodes(self, nodes: list):
        """
        Remove specified nodes from the reservoir network.
        Parameters:
        nodes (list): A list of indices representing the nodes to be removed.
        1. Removes the specified nodes from the adjacency matrix.
        2. Updates the reservoir properties including the number of nodes, density, and spectral radius.
        """
        # remove nodes from the reservoir network

        if not isinstance(nodes, list):
            raise TypeError("Nodes must be provided as a list of indices.")

        if np.max(nodes) > self.nodes:
            raise ValueError("Node index exceeds the number of nodes in the reservoir.")

        if np.min(nodes) < 0:
            raise ValueError("Node index must be positive.")

        # 1. remove nodes from the adjacency matrix
        self.weights = remove_nodes_from_graph(graph=self.weights, nodes=nodes)

        # 2. update reservoir properties
        self.update_layer_properties()

        # 3. update the initial reservoir state
        self.initial_res_states = np.delete(self.initial_res_states, nodes, axis=0)

        # # 4. update the info about input-receiving nodes
        # self.input_receiving_nodes = np.delete(self.input_receiving_nodes, nodes)

    def update_layer_properties(self):
        """
        Updates the reservoir properties including the number of nodes, density, and spectral radius.
        This method should be called after any changes to the reservoir network.
        """
        self.nodes = get_num_nodes(self.weights)
        self.density = compute_density(self.weights)
        self.spec_rad = compute_spec_rad(self.weights)

    def get_spec_rad(self):
        return self.spec_rad

    def set_spec_rad(self, value):
        """ Sets the spectral radius of the reservoir network.
        Parameters:
        - value (float): The desired spectral radius.
        Raises:
        - ValueError: If the value is not positive or if the weights are not set."""

        # check that the value is positive and that the weights are set
        if value <= 0:
            raise ValueError("Spectral radius must be positive")
        if self.weights is not None:
            self.weights = set_spec_rad(self.weights, value)
        else:
            raise ValueError("Weights must be set before changing spectral radius.")
        
        # check that the weights re-scaling was successful
        _new_spec_rad = compute_spec_rad(self.weights)
        if np.isclose(value, _new_spec_rad, atol=1e-5):
            self.spec_rad = _new_spec_rad
        else:
            raise ValueError(
                f"Failed to set spectral radius within tolerance 1e-5. Expected {value}, got {_new_spec_rad}."
            )

    def get_leakage_rate(self):
        return self.leakage_rate

    def set_leakage_rate(self, value):
        if not (0 < value <= 1):
            raise ValueError("Leakage rate must be in (0, 1]")
        self.leakage_rate = value

    def get_activation(self):
        return self.activation

    def set_activation(self, value):
        if value not in ["tanh", "sigmoid"]:
            raise ValueError("Activation must be 'tanh' or 'sigmoid'")
        self.activation = value


class RandomReservoirLayer(ReservoirLayer):
    def __init__(
        self,
        nodes,
        density: float = 0.1,
        activation: str = "tanh",
        leakage_rate: float = 0.5,
        fraction_input: float = 0.8,
        spec_rad: float = 0.9,
        init_res_sampling="random_normal",
        seed=None,
    ):

        # Call the parent class's __init__ method
        super().__init__(
            nodes=nodes,
            density=density,
            activation=activation,
            leakage_rate=leakage_rate,
            fraction_input=fraction_input,
            init_res_sampling=init_res_sampling,
            seed=seed,
        )

        # initialize subclass-specific attributes
        self.seed = seed
        self.spec_rad = spec_rad

        # generate a random ER graph using networkx
        self.weights = gen_ER_graph(
            nodes=nodes,
            density=density,
            spec_rad=self.spec_rad,
            directed=True,
            seed=seed,
        )

        # update the spectral radius and the density to the actual values
        self.update_layer_properties()

        # check that the graph actually satisfies the spectral radius and 
        # density requested by the user
        _spec_rad = compute_spec_rad(self.weights)
        _density = compute_density(self.weights)
        if np.isclose(self.spec_rad, _spec_rad, atol=1e-5):
            self.spec_rad = _spec_rad
        else:
            raise ValueError(
                f"Failed to set spectral radius within tolerance 1e-5. Expected {self.spec_rad}, got {_spec_rad}."
            )
        
        if np.isclose(self.density, _density, atol=1e-5):
            self.density = _density
        else:
            raise ValueError(
                f"Failed to set density within tolerance 1e-5. Expected {self.density}, got {_density}."
            )

    #def update_layer_properties(self):
        #Updates the reservoir properties including the number of nodes, density, and spectral radius.
        #self.nodes = get_num_nodes(self.weights)
        #self.density = compute_density(self.weights)
        # self.spec_rad = compute_spec_rad(self.weights)


# class ReccurrenceLayer(ReservoirLayer):
#     # To Do: accept a random seed
#     def __init__(self, nodes, density, activation: str = 'tanh', leakage_rate: float = 0.2):
#         # Call the parent class's __init__ method
#         super().__init__(nodes, density, activation, leakage_rate)
#
#         # Initialize subclass-specific attributes
#         # https://pyts.readthedocs.io/en/stable/generated/pyts.image.RecurrencePlot.html#pyts.image.RecurrencePlot
#         # https://tocsy.pik-potsdam.de/pyunicorn.php
#



"""
Capabilities to prune an existing RC model, i.e. try to cut reservoir nodes and improve 
performance while reducing the reservoir size
"""

import numpy as np
from pyreco.custom_models import RC
from pyreco.node_selector import NodeSelector
import networkx as nx
import math
from typing import Union
import copy
from pyreco.graph_analyzer import GraphAnalyzer
from pyreco.node_analyzer import NodeAnalyzer


class NetworkPruner:
    # implements a pruning object for pyreco objects.

    def __init__(
        self,
        target_score: float = None,
        stop_at_minimum: bool = True,
        min_num_nodes: int = 3,
        patience: int = 0,
        candidate_fraction: float = 0.1,
        remove_isolated_nodes: bool = False,
        criterion: str = "mse",
        metrics: Union[list, str] = ["mse"],
        maintain_spectral_radius: bool = False,
        node_props_extractor=None,
        graph_props_extractor=None,
        return_best_model: bool = True,
        graph_analyzer: GraphAnalyzer = None,
        node_analyzer: NodeAnalyzer = None,
    ):
        """
        Initializer for the pruning class.

        Parameters:
        - target_score (float): The test set score that the user aims at. Pruning stops
        once this score is reached.

        - stop_at_minimum (bool): Whether to stop at the local minimum of the test set
        score. When set to False, pruning continues until the minimal number of nodes
        in <min_num_nodes>.

        - min_num_nodes (int): Stop pruning when arriving at this number of nodes.
        Conflicts if stop_at_minimum is set to True but also a min_num_nodes is given.

        - patience (int): We allow a patience, i.e. keep pruning after we reached a
        (local) minimum of the test set score. Depends on the size of the original
        reservoir network, defaults to 10% of initial reservoir nodes.

        - candidate_fraction (float): number of randomly chosen reservoir nodes during
        every pruning iteration that is a candidate for pruning. Refers to the fraction of nodes w.r.t. current number of nodes during pruning iteration.

        - remove_isolated_nodes (bool): Whether to remove isolated nodes during pruning.

        - criterion (str): The criterion to be used for steering the node pruning. Default is "mse".

        - metrics (list or str): The metrics to be used for evaluating the pruned model. Default is ["mse"].

        - maintain_spectral_radius (bool): Whether to maintain the spectral radius of the reservoir layer during pruning.
        """

        # Sanity checks for the input parameter types and values
        if target_score is not None and not isinstance(target_score, float):
            raise TypeError("target_score must be a float")

        if not isinstance(stop_at_minimum, bool):
            raise TypeError("stop_at_minimum must be a boolean")

        if not isinstance(min_num_nodes, int):
            raise TypeError("min_num_nodes must be an integer")
        if min_num_nodes <= 2:
            raise ValueError("min_num_nodes must be larger than 2")
        if patience is not None and not isinstance(patience, int):
            raise TypeError("patience must be an integer")

        if not isinstance(candidate_fraction, float):
            raise TypeError("candidate_fraction must be a float in (0, 1]")

        if candidate_fraction <= 0 or candidate_fraction > 1:
            raise ValueError("candidate_fraction must be a float in (0, 1]")

        if not isinstance(criterion, str):
            raise TypeError("criterion must be a string")

        if graph_analyzer is not None and not isinstance(graph_analyzer, GraphAnalyzer):
            raise TypeError("graph_analyzer must be an instance of GraphAnalyzer")
        if graph_analyzer is None:
            graph_analyzer = GraphAnalyzer()

        if node_analyzer is not None and not isinstance(node_analyzer, NodeAnalyzer):
            raise TypeError("node_analyzer must be an instance of NodeAnalyzer")
        if node_analyzer is None:
            node_analyzer = NodeAnalyzer()

        # Assigning the parameters to instance variables
        if target_score is None:
            self.target_score = 0.0
        else:
            self.target_score = target_score

        self.criterion = criterion
        self.stop_at_minimum = stop_at_minimum
        self.min_num_nodes = min_num_nodes
        self.patience = patience
        self.candidate_fraction = candidate_fraction
        self.metrics = metrics
        self.return_best_model = return_best_model
        self.graph_analyzer = graph_analyzer
        self.node_analyzer = node_analyzer

        # TODO not implemented yet
        self.remove_isolated_nodes = remove_isolated_nodes
        self.maintain_spectral_radius = maintain_spectral_radius

        # store the history of the pruning process in a nested dictionary
        self.history = {}

        # initialize attributes that will be used during pruning (and changed during the process)
        # needs to be attributes as the history updates depend on them
        self._curr_loss = None
        self._curr_num_nodes = None
        self._curr_loss_history = []
        self._idx_prune = None
        self._patience_counter = 0
        self._curr_metrics = None

    def prune(self, model: RC, data_train: tuple, data_val: tuple):
        """
        Prune a given model by removing nodes.

        Parameters:
        - model (RC): The reservoir computer model to prune.
        - data_train (tuple): Training data.
        - data_val (tuple): Validation data.
        """

        # Sanity checks for the input parameter types and values
        if not isinstance(model, RC):
            raise TypeError("model must be an instance of RC")

        if not isinstance(data_train, tuple) or not isinstance(data_val, tuple):
            raise TypeError("data_train and data_val must be tuples")

        if len(data_train) != 2 or len(data_val) != 2:
            raise ValueError("data_train and data_val must have 2 elements each")

        for idx, elem in enumerate(data_train):
            if not isinstance(elem, list):
                if not isinstance(elem, np.ndarray):
                    raise TypeError(f"data_train[{idx}] must be a list or numpy array")

        for idx, elem in enumerate(data_val):
            if not isinstance(elem, list):
                if not isinstance(elem, np.ndarray):
                    raise TypeError(f"data_val[{idx}] must be a list or numpy array")

        if len(data_train[0]) != len(data_train[1]):
            raise ValueError(
                "data_train[0] and data_train[1] must have the same length, "
                "i.e. same number of samples"
            )

        if len(data_val[0]) != len(data_val[1]):
            raise ValueError(
                "data_val[0] and data_val[1] must have the same length, "
                "i.e. same number of samples"
            )

        # obtain training and testing data
        x_test, y_test = data_val[0], data_val[1]
        x_train, y_train = data_train[0], data_train[1]

        # Assigning the parameters to instance variables that can not be set
        # in the initializer, as they depend on the model and data
        self._curr_num_nodes = model.reservoir_layer.nodes

        # initialize the quantities that affect the stop condition
        self._curr_loss = model.evaluate(x=x_test, y=y_test, metrics=self.criterion)[0]
        self._curr_loss_history = [self._curr_loss]

        # initialize quantities that we track for the pruning history
        # these do not affect the pruning process
        self._curr_metrics = model.evaluate(x=x_test, y=y_test, metrics=self.metrics)

        # storing all pruned models during the pruning iteration
        # allows to recover models from previous iterations, e.g. when the best model is not the last one in the iteration (positive patience value)
        _pruned_models = [copy.deepcopy(model)]

        # initialize the pruning iterator
        self._iter_count = 0

        # Store all relevant information during pruning inside self.history
        # self._update_pruning_history(model=model)
        self.add_val_to_history(["loss"], self._curr_loss)
        self.add_val_to_history(["metrics"], self._curr_metrics)
        self.add_val_to_history(["num_nodes"], self._curr_num_nodes)
        self.add_val_to_history(["iteration"], self._iter_count)

        _graph = model.reservoir_layer.weights
        _graph_props = self.graph_analyzer.extract_properties(graph=_graph)
        self.add_dict_to_history(["graph_props"], _graph_props)

        while True:  # self._curr_num_nodes>self.min_num_nodes:

            print(f"pruning iteration {self._iter_count}")

            print(
                f"current reservoir size: {self._curr_num_nodes}, current loss: {self._curr_loss:.8f}"
            )

            # propose a list of nodes to prune using a random uniform distribution. If the user specified a candidate_fraction of 1.0, we will try out all nodes
            _num_nodes_to_prune = math.ceil(
                self.candidate_fraction * self._curr_num_nodes
            )
            selector = NodeSelector(
                total_nodes=self._curr_num_nodes, strategy="random_uniform_wo_repl"
            )
            # obtain nodes that are proposed for pruning
            _curr_candidate_nodes = selector.select_nodes(num=_num_nodes_to_prune)
            print(
                f"propose {_num_nodes_to_prune}/{self._curr_num_nodes} nodes for pruning"
            )

            # track the performance of the RC with the candidate nodes removed
            _candidate_scores = []
            _candidate_models = []
            _cand_node_props = []  # properties of the to-be removed node
            _cand_node_input_receiving = (
                []
            )  # whether the node is connected to input layer
            _cand_node_output_sending = (
                []
            )  # whether the node is connected to output layer
            _cand_graph_props_before = []  # properties of the graph before pruning
            _cand_graph_props_after = []  # properties of the graph after pruning

            # iteratate over the candidate nodes: delete one-by-one, measure performance,
            # and also track node/graph-level properties
            for node in _curr_candidate_nodes:

                # get a copy of the original model to try out the deletion
                _model = copy.deepcopy(model)

                # extract information about the node that we will prune,
                # and about the graph before we prune it
                _graph = _model.reservoir_layer.weights
                _node_props = self.node_analyzer.extract_properties(
                    graph=_graph, node=node
                )
                _graph_props = self.graph_analyzer.extract_properties(graph=_graph)

                # check for links to input and read-out layer of the current node
                _is_input_receiving = (
                    node in _model.reservoir_layer.input_receiving_nodes
                )
                _is_output_sending = node in _model.readout_layer.readout_nodes

                _cand_node_props.append(_node_props)
                _cand_graph_props_before.append(_graph_props)
                _cand_node_input_receiving.append(_is_input_receiving)
                _cand_node_output_sending.append(_is_output_sending)

                # remove current candidate node
                _model.remove_reservoir_nodes(nodes=[node])

                # TODO: remove isolated nodes using utility function from utils_networks
                # if self.remove_isolated_nodes:
                # iso_nodes = ...
                # _model.remove_reservoir_nodes(nodes=[iso_nodes])

                # TODO: maintain the spectral radius of the reservoir layer
                # if self.maintain_spectral_radius:
                # spec_rad = model.reservoir_layer.spectral_radius
                # _model.set_spec_rad(spec_rad)

                # pruning requires re-fitting the model
                _model.fit(x=x_train, y=y_train)

                # evaluate the pruned model
                _score = _model.evaluate(x=x_test, y=y_test, metrics=self.criterion)[0]

                # extract graph properties after pruning
                _graph = _model.reservoir_layer.weights
                _graph_props = self.graph_analyzer.extract_properties(graph=_graph)
                _cand_graph_props_after.append(_graph_props)

                print(
                    f"deletion of candidate node {node}. loss: \t{_score:.6f} ({(self._curr_loss-_score)/self._curr_loss:+.3%})"
                )

                # store the relevant candidate information
                _candidate_scores.append(_score)
                _candidate_models.append(_model)

                # delete temporary variables (just for safety)
                del (
                    _model,
                    _score,
                    _graph,
                    _graph_props,
                    _node_props,
                )

            # store the candidate properties in the history object
            self.add_val_to_history(
                ["candidate_scores"],
                _candidate_scores,
            )

            self.add_val_to_history(
                ["candidate_nodes"],
                _curr_candidate_nodes,
            )

            self.add_val_to_history(
                ["candidate_node_props"],
                dictlist_to_dict(_cand_node_props),
            )

            self.add_val_to_history(
                ["candidate_graph_props_before"],
                dictlist_to_dict(_cand_graph_props_before),
            )

            self.add_val_to_history(
                ["candidate_graph_props_after"],
                dictlist_to_dict(_cand_graph_props_after),
            )

            # after trying out all candidate nodes, we need to select the node to prune,
            # i.e. the one that has the smallest loss among all candidate nodes
            idx_prune = np.argmin(_candidate_scores)
            self._curr_idx_prune = idx_prune  # just for history logging

            # update the termination relevant quantities,
            # assuming that we will prune that node
            self._curr_loss = _candidate_scores[idx_prune]
            self._curr_num_nodes = _candidate_models[idx_prune].reservoir_layer.nodes
            self._curr_loss_history.append(self._curr_loss)

            # check if we should actually prune the node, or if that would violate the termination criteria (no optimal design by now to do it here though)
            if not self._keep_pruning():

                # exit the pruning loop
                break

            print(f"pruning node {idx_prune}, resulting in loss {self._curr_loss:.6f}")
            print(
                f"loss improvement by {((self._curr_loss_history[-2]-self._curr_loss)/self._curr_loss_history[-2]):+.3%}\n"
            )

            # prune the node that gives us the least performance drop. as we have already
            # pruned the node and stored the model, we only need to update the model.
            # Saves at least one training run and all the pruning logic
            model = _candidate_models[idx_prune]

            # store the model for later use
            _pruned_models.append(copy.deepcopy(model))

            # compute things that are required for the history, but not for the
            # pruning loop termination criteria
            self._curr_metrics = model.evaluate(
                x=x_test, y=y_test, metrics=self.metrics
            )

            # self._update_pruning_history

            self.add_val_to_history(["loss"], self._curr_loss)
            self.add_val_to_history(["metrics"], self._curr_metrics)
            self.add_val_to_history(["num_nodes"], self._curr_num_nodes)
            self.add_val_to_history(["idx_prune"], self._curr_idx_prune)
            self.add_val_to_history(["iteration"], self._iter_count)

            self.add_val_to_history(
                ["del_node_props", "input_receiving_node"],
                _cand_node_input_receiving[idx_prune],
            )
            self.add_val_to_history(
                ["del_node_props", "output_sending_node"],
                _cand_node_output_sending[idx_prune],
            )

            self.add_dict_to_history(["del_node_props"], _cand_node_props[idx_prune])
            self.add_dict_to_history(
                ["graph_props"], _cand_graph_props_after[idx_prune]
            )

            # update counter
            self._iter_count += 1

        # in case we have a non-zero patience, we need to return the best model
        # instead of the last one (i.e. when a positive patience value was given)
        if self.return_best_model:
            idx_best = np.argmin(self._curr_loss_history[:-1])
            model = copy.deepcopy(_pruned_models[idx_best])
            print(f"returning model {idx_best} as the best, i.e. with lowest loss")

        # we should fit the final model, and evaluate it
        model.fit(x=x_train, y=y_train)
        final_loss = model.evaluate(x=x_test, y=y_test, metrics=self.criterion)[0]
        final_metrics = model.evaluate(x=x_test, y=y_test, metrics=self.metrics)
        print(
            f"\ninitial loss{self._curr_loss_history[0]:.6f}, loss after pruning: {final_loss:.6f}"
        )
        print(f"final model has {model.reservoir_layer.nodes} nodes")
        print(f"final model loss {self.criterion}: {final_loss:.6f}")
        print(f"final model metrics ({self.metrics}): {final_metrics}")
        return model, self.history

    def _keep_pruning(self):
        # Termination criteria for the pruning process

        # Keep pruning as long as all of the following conditions are met:
        # 1. The current score is below the target score
        # 2. The current number of nodes is above the minimum number of nodes
        # 3. The current loss is smaller than the previous loss

        if (
            self._loss_not_met()
            and self._num_nodes_not_met()
            and self._at_minimum_not_met()
        ):
            return True
        else:
            return False

    def _loss_not_met(self):
        # Checks the stopping condition based on the current and target loss
        # returns True if the criterion is not met, i.e. we should continue pruning
        if self._curr_loss >= self.target_score:
            print(
                f"Loss {self._curr_loss:.6f} is larger target score {self.target_score:.6f}. Continuing pruning"
            )
            return True
        else:
            print(
                f"Loss {self._curr_loss:.6f} is smaller target score {self.target_score:.6f}. Terminating pruning"
            )
            return False

    def _num_nodes_not_met(self):
        # checks if the number of nodes is above the minimum number of nodes
        # returns True if number of nodes is above minimum, i.e. we should continue pruning
        if self._curr_num_nodes > self.min_num_nodes:
            print(
                f"Number of nodes {self._curr_num_nodes} is larger than minimum number of nodes {self.min_num_nodes}. Continuing pruning"
            )
            return True
        else:
            print(
                f"Number of nodes {self._curr_num_nodes} is smaller/equal minimum number of nodes {self.min_num_nodes}. Terminating pruning"
            )
            return False

    def _at_minimum_not_met(self):
        # checks if the loss is at a minimum,
        # considering also patience.
        # returns True if loss is not at minimum, i.e. we should continue pruning
        if len(self._curr_loss_history) < 2:
            # we are just at the start of pruning, cannot
            # check for a minimum.
            return True

        if self.stop_at_minimum:
            if self._curr_loss_history[-2] > self._curr_loss_history[-1]:
                print(
                    f"Loss decreased from {self._curr_loss_history[-2]:.6f} to {self._curr_loss_history[-1]:.6f}. Continuing pruning"
                )
                self._patience_counter = 0
                return True
            else:  # current loss is larger than previous
                self._patience_counter += 1
                if self._patience_counter < self.patience:
                    print(
                        f"Loss increased, but {self._patience_counter} < {self.patience} Continuing pruning"
                    )
                    return True
                else:
                    # TODO: we need to recover the model that had the best score!
                    print(
                        f"Loss increased for {self.patience} consecutive iterations. Terminating pruning"
                    )
                    return False
        else:
            return True

    def add_val_to_history(self, keys, value):
        """
        Add a value to history dictionary based on a list of keys.

        Args:
            keys (list): A list of keys specifying the path in the nested dictionary.
            value: The value to add.
        """
        if len(keys) == 1:
            if keys[0] not in self.history:
                self.history[keys[0]] = []
            self.history[keys[0]].append(value)

        elif len(keys) == 2:
            if keys[0] not in self.history:
                self.history[keys[0]] = {}
            if keys[1] not in self.history[keys[0]]:
                self.history[keys[0]][keys[1]] = []
            self.history[keys[0]][keys[1]].append(value)

        # for key in keys[:-1]:
        #     if key not in self.history:
        #         self.history[key] = {}
        #     self.history = self.history[key]
        # if keys[-1] not in self.history:
        #     self.history[keys[-1]] = []
        # self.history[keys[-1]].append(value)

    def add_dict_to_history(self, keys, value_dict):
        """
        Add a dictionary to history dictionary based on a list of keys.

        Args:
        nested_dict (dict): The nested dictionary.
        keys (list): A list of keys specifying the path in the nested dictionary.
        value_dict (dict): The dictionary to add.
        """

        for key in keys[:-1]:
            if key not in self.history:
                self.history[key] = {}
            self.history = self.history[key]
        if keys[-1] not in self.history:
            self.history[keys[-1]] = {}
        for k, v in value_dict.items():
            if k not in self.history[keys[-1]]:
                self.history[keys[-1]][k] = []
            self.history[keys[-1]][k].append(v)


def dictlist_to_dict(dict_list):
    """
    Join dictionaries in a list into a common dictionary.

    Args:
        dict_list (list): A list of dictionaries to join.

    Returns:
        dict: A common dictionary containing all key-value pairs from the dictionaries in the list.
    """
    common_dict = {}
    for d in dict_list:
        for key, value in d.items():
            if key in common_dict:
                if isinstance(common_dict[key], list):
                    common_dict[key].append(value)
                else:
                    common_dict[key] = [common_dict[key], value]
            else:
                common_dict[key] = value
    return common_dict

    # def _update_pruning_history(self, model: RC):
    #     # this will keep track of all quantities that are relevant during the pruning iterations.

    #     # Pruning iteration
    #     # self.history["iteration"].append(self._curr_iter)

    #     if not self.history:
    #         # initialize the history object
    #         self.history["iteration"] = []
    #         self.history["loss"] = []
    #         self.history["metrics"] = []
    #         self.history["num_nodes"] = []

    #         # initialize the dicts for the graph and node properties with empty lists
    #         graph_keys = self.graph_analyzer.list_properties()
    #         node_keys = self.node_analyzer.list_properties()

    #         self.history["graph_props"] = {key: [] for key in graph_keys}
    #         # self.history["candidate_graph_props"] = {key: [] for key in graph_keys}

    #         self.history["del_node_props"] = {key: [] for key in graph_keys}
    #         # self.history["candidate_node_props"] = {key: [] for key in graph_keys}

    #     else:
    #         # store the most relevant information

    #         # we will extract properties from the reservoir network of the model
    #         graph = model.reservoir_layer.weights
    #         graph_props = self.graph_analyzer.extract_properties(graph)

    #         # # choose the node to extract properties from
    #         # node = int(self._curr_idx_prune)
    #         # node_props = self.node_analyzer.extract_properties(graph, node)

    #         # high-level properties
    #         self.add_val_to_history(
    #             ["num_nodes"],
    #             self._curr_num_nodes,
    #         )

    #         self.add_val_to_history(
    #             ["loss"],
    #             self._curr_loss,
    #         )

    #         self.add_val_to_history(
    #             ["metrics"],
    #             self._curr_metrics,
    #         )

    #         self.add_val_to_history(
    #             ["iteration"],
    #             self._iter_count,
    #         )
    #         self.add_val_to_history(
    #             ["graph_props"],
    #             graph_props,
    #         )


def append_to_dict(dict1, dict2):
    # appends entries in dict1 to existing dict 2

    for key in list(dict1.keys()):
        if key in list(dict2.keys()):
            # print(f"appending {key} to existing dict")
            dict2[key].append(dict1[key])

    return dict2


if __name__ == "__main__":
    # test the pruning

    from pyreco.utils_data import sequence_to_sequence as seq_2_seq
    from pyreco.custom_models import RC as RC
    from pyreco.layers import InputLayer, ReadoutLayer
    from pyreco.layers import RandomReservoirLayer
    from pyreco.optimizers import RidgeSK

    # get some data
    X_train, X_test, y_train, y_test = seq_2_seq(
        name="sine_pred", n_batch=20, n_states=2, n_time=150
    )

    input_shape = X_train.shape[1:]
    output_shape = y_train.shape[1:]

    # build a classical RC
    model = RC()
    model.add(InputLayer(input_shape=input_shape))
    model.add(
        RandomReservoirLayer(
            nodes=50,
            density=0.1,
            activation="tanh",
            leakage_rate=0.1,
            fraction_input=0.5,
        ),
    )
    model.add(ReadoutLayer(output_shape, fraction_out=0.9))

    # Compile the model
    optim = RidgeSK(alpha=0.5)
    model.compile(
        optimizer=optim,
        metrics=["mean_squared_error"],
    )

    # Train the model
    model.fit(X_train, y_train)

    print(f"score: \t\t\t{model.evaluate(x=X_test, y=y_test)[0]:.4f}")

    # prune the model
    pruner = NetworkPruner(
        stop_at_minimum=False,
        min_num_nodes=46,
        patience=2,
        candidate_fraction=0.9,
        remove_isolated_nodes=False,
        metrics=["mse"],
        maintain_spectral_radius=False,
    )

    model_pruned, history = pruner.prune(
        model=model, data_train=(X_train, y_train), data_val=(X_test, y_test)
    )

    import matplotlib.pyplot as plt

    plt.figure()
    plt.subplot(1, 2, 1)
    plt.plot(history["num_nodes"], history["loss"], label="loss")
    plt.xlabel("number of nodes")
    plt.ylabel("loss")
    plt.subplot(1, 2, 2)
    for key in history["graph_props"].keys():
        plt.plot(history["num_nodes"], history["graph_props"][key], label=key)
    plt.xlabel("number of nodes")
    plt.yscale("log")
    plt.legend()
    plt.show()



"""
Some testing use case for the CustomModel API of pyreco
"""

import numpy as np
from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.utils_data import sequence_to_scalar
from pyreco.optimizers import RidgeSK

"""
Classic RC built on random networks. 

Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal from a sequence of 20 function values
X_train, X_test, y_train, y_test = sequence_to_scalar(
    name="sine_prediction",
    n_states=1,
    n_batch=200,
    n_time_in=20,
)

# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(
    RandomReservoirLayer(
        nodes=200, density=0.1, activation="tanh", leakage_rate=0.1, fraction_input=0.5
    ),
)
model_rc.add(ReadoutLayer(output_shape, fraction_out=0.99))

# Compile the model
optim = RidgeSK(alpha=0.5)
model_rc.compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
)

# Train the model
model_rc.fit(X_train, y_train)

# Make predictions for new data
y_pred = model_rc.predict(X_test)
print(f"shape of predictions on test set: {y_pred.shape}")

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=["mae"])
print(f"\nTest model loss (old input weights): {loss_rc[0]:.6f}")


# ""
# Now re-set the read-in weights
# ""

# check _connect_input_to_reservoir(self) in custom_models.py
# to see how we set the input weights.
new_input_weights = np.random.randn(200, 1)


#
# model_rc.input_layer.weights = new_input_weights  # works as well
model_rc._set_readin_weights(new_input_weights)

# Train the model
model_rc.fit(X_train, y_train)

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=["mae"])
print(f"\nTest model loss (new input weights): {loss_rc[0]:.6f}")



# test the pruning
import numpy as np
import matplotlib.pyplot as plt
from pyreco.utils_data import sequence_to_sequence as seq_2_seq
from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.optimizers import RidgeSK


# get some data
X_train, X_test, y_train, y_test = seq_2_seq(
    name="sine_pred", n_batch=20, n_states=2, n_time=150
)

input_shape = X_train.shape[1:]
output_shape = y_train.shape[1:]

# build a classical RC
model = RC()
model.add(InputLayer(input_shape=input_shape))
model.add(
    RandomReservoirLayer(
        nodes=50,
        density=0.1,
        activation="tanh",
        leakage_rate=0.1,
        fraction_input=1.0,
    ),
)
model.add(ReadoutLayer(output_shape, fraction_out=0.9))

# Compile the model
optim = RidgeSK(alpha=0.5)
model.compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
)

# Train the model
model.fit(X_train, y_train)

print(f"score: \t\t\t{model.evaluate(x=X_test, y=y_test)[0]:.4f}")


"""
Pruning Part
"""
from pyreco.pruning import NetworkPruner

# prune the model
pruner = NetworkPruner(
    stop_at_minimum=True,
    min_num_nodes=20,
    patience=2,
    candidate_fraction=0.1,  # 1.0 would try out every possible node
    remove_isolated_nodes=False,
    metrics=["mse"],
    maintain_spectral_radius=False,
    return_best_model=True,
)

model_pruned, history = pruner.prune(
    model=model, data_train=(X_train, y_train), data_val=(X_test, y_test)
)

print(f"took {history['iteration'][-1]+1} iterations to prune the model")
for key in history["graph_props"].keys():
    print(
        f"{key}: \t initial model {history['graph_props'][key][0]:.4f}; \t final model: {history['graph_props'][key][-1]:.4f}"
    )

plt.figure()
plt.subplot(1, 2, 1)
plt.plot(history["num_nodes"], history["loss"], label="loss")
plt.xlabel("number of nodes")
plt.ylabel("loss")
plt.subplot(1, 2, 2)
for key in history["graph_props"].keys():
    plt.plot(history["num_nodes"], history["graph_props"][key], label=key)
plt.xlabel("number of nodes")
plt.yscale("log")
plt.legend()
plt.show()


# investigate a single decision: which node was pruned, and which pruning candidate properties do we have?
iteration = 2
plt.figure()
plt.hist(x=history["candidate_node_props"][iteration]["degree"])
plt.title(
    f'pruned node degree was {history["del_node_props"]["degree"][iteration]} in iteration {iteration}'
)
plt.legend(["pruning candidate nodes"])
plt.xlabel("degree")
plt.show()



"""
Some testing use case for the CustomModel API of pyreco
"""

from re import X
from matplotlib import pyplot as plt
from pyreco.custom_models import AutoRC as RC
from pyreco.layers import InputLayer, ReadoutLayer, FeedbackLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.plotting import r2_scatter
from pyreco.utils_data import sequence_to_sequence
from pyreco.optimizers import RidgeSK

from scipy.integrate import solve_ivp
import numpy as np
"""
Classic RC built on random networks. 
Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal from a sequence of 20 function values
A_train, A_test, B_train, B_test = sequence_to_sequence(
    name="sine_pred",
    n_states=1,
    n_batch=50,
    n_time=201
)


#####testing hybrid RC model on forced Duffing oscillator data#############################
def Plot(a, b, forcing):
    fig_size = plt.rcParams["figure.figsize"]  
    fig_size[0] = 6; fig_size[1] = 1.5
    plt.rcParams["figure.figsize"] = fig_size 
    plt.figure()
    plt.subplot(3, 1, 1)
    plt.plot(a[:1000], lw=0.5, color='black')
    plt.ylabel(r'$q_1(t)$')
    plt.title('damped forced oscillator')

    plt.subplot(3, 1, 2)
    plt.plot(b[:1000], lw=0.5, color='black')
    plt.ylabel(r'$q_2(t)$')

    plt.subplot(3, 1, 3)
    plt.plot(forcing[:1000], lw=0.5, color='red')
    plt.xlabel(r'$t$')
    plt.ylabel(r'$f(t)$')
    plt.tight_layout()
    plt.show()

    fig_size = plt.rcParams["figure.figsize"]  
    fig_size[0] = 1.5; fig_size[1] = 1.5
    plt.rcParams["figure.figsize"] = fig_size 
    plt.plot(a, b, lw=0.5)
    plt.scatter(a[0], b[0], marker='o',c='r')
    plt.xlabel(r'$q_1(t)$'); plt.ylabel(r'$q_2(t)$')
    plt.show()

def one_dof_oscillator(t, q: np.ndarray,
                       c: float = 0.1,
                       k: float = 1.0,
                       f: float = 0,
                       omega: float = 1.0,
                       phi: float = 0.0,
                       beta: float=0
                       ) -> np.ndarray:
    """ ODE one-dimensional oscillator. """

    A = np.array([[0, 1], [-k, -c]])
    B = np.array([[0, 0], [-beta, 0]])
    F = np.array([0, f*np.cos(omega*t+phi)])

    return np.dot(A, q) + np.dot(B, q*q*q) + F

def SolveDuffing(q0, t_eval, c, k, f, omega, phi, beta):
    # numerical time integration
    sol_forced = solve_ivp(one_dof_oscillator, t_span=[t_eval[0], t_eval[-1]], y0=q0,\
                           t_eval=t_eval, args=(c, k, f, omega, phi, beta))

    # display of trajectories
    q = sol_forced.y.T
    t = sol_forced.t
    forcing = f*np.cos(omega*t+phi)  
    # Plot(q[:, 0], q[:, 1], forcing) 
    return q, forcing

def InputGenerator(t_eval,Trans,c,k,f,omega,phi,beta,q0):
    Q_i=[];Forcing_i=[]
    for i in range(len(f)):
        for j in range(len(omega)):
            Qs, Fs = SolveDuffing(q0, t_eval, c, k, f[i], omega[j], phi, beta)
            Q_i.append(Qs); Forcing_i.append(Fs)
    
    Q_i=np.asarray(Q_i); Forcing_i=np.asarray(Forcing_i)   
    
    #####Input##################
    Inp0=[]; Inp1=[]; Inp2=[]
    for i in range(len(f)*len(omega)):
        Inp0.append(Forcing_i[i,Trans:])
        Inp1.append(Q_i[i,Trans:,0])
        Inp2.append(Q_i[i,Trans:,1])
    
    #####Plot##################
    for i in range(len(f)*len(omega)):
        Plot(Inp1[i], Inp2[i], Inp0[i])

    return np.array([np.asarray(Inp0), np.asarray(Inp1), np.asarray(Inp2)])

# ######### Duffing Oscillator parameters####################################
c = 0.3 ###0.32    # damping
k = -1.0     # linear stiffness
f = [0.49] #[0.475, 0.485] #[0.47, 0.49]    # forced case
omega = [1.46] #[1.5, 1.53]  #[1.5, 1.53]  ####good in [1.47, 1.59] for 1st bif. pt.
phi = 0
beta=1

#### initial conditions
q0 = np.array([0.05, 0.05])

#### time integration interval
T=6000; h=0.1; Trans=500
t_eval = np.arange(start=0, stop=T, step=h)

Inps = InputGenerator(t_eval,Trans,c,k,f,omega,phi,beta,q0)
print('Input shapes: ',Inps[:,0,:].T.shape)

TPts = int((T/h)-Trans)
Batch = 25

Inps_N = np.reshape(Inps[:,0,:].T, (Batch, int(TPts/Batch),  3))
print('Input reshaped to [batch, time, states]: ',Inps_N.shape)

X_train = Inps_N[:20, :, :]  # use first 8 samples for training
X_test = Inps_N[20:, :, :]   # use last 2 samples for testing

y_train = Inps_N[:20, :, 1:]  # use first 8 samples for training, only q1 and q2 as target
y_test = Inps_N[20:, :, 1:]   # use last 2 samples for testing, only q1 and q2 as target

# print('Duffing Training data shape: ',D_train.shape)
# print('Duffing Test data shape: ',D_test.shape)
#######################################################################################

###generate y_train and y_test as previous time step of the input series
# X_train = A_train#[:, 1:, :]
# X_test = A_test#[:, 1:, :]
# ### remove last time step from X_train and X_test
# y_train = A_train#[:, :-1, :]
# y_test = A_test#[:, :-1, :]

print(f"Generated training data shapes: {X_train.shape}, {y_train.shape}")
print(f"Generated test data shapes: {X_test.shape}, {y_test.shape}")

# ##plot x and y
# plt.figure()
# for i in range(X_train.shape[0]):
#     plt.subplot(X_train.shape[0], 1, i + 1)
#     plt.plot(X_train[i, :, 0], color="k", alpha=0.3)
#     plt.plot(y_train[i, :, 0], color="b", alpha=0.3)
#     plt.xlabel("Time")
#     plt.ylabel("Amplitude")
# plt.tight_layout()
# plt.show()

# print(f"X_train shape: {X_train.shape}, y_train shape: {y_train.shape}")
# print(f"X_test shape: {X_test.shape}, y_test shape: {y_test.shape}")


# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])


model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(FeedbackLayer(feedback_shape=output_shape))
model_rc.add(
    RandomReservoirLayer(
        nodes=100, density=0.5, spec_rad=0.5057, activation="tanh", leakage_rate=0.2, fraction_input=0.25),
)
model_rc.add(ReadoutLayer(output_shape, fraction_out=1.0))

# Compile the model
optim = RidgeSK(alpha=5e-7) #1.0)
model_rc.AutoRC_compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
    discard_transients=10
)

# Train the model
model_rc.fit(X_train, y_train)

# print(model_rc.input_layer.weights)
# Make predictions for new data
y_pred = model_rc.predict(X_test)

###plot FFRC predictions
plt.figure()
plt.title("Single-step predictions. No feedback connections.")
for i in range(y_test.shape[0]):
    for j in range(y_test.shape[2]):
        plt.subplot(y_test.shape[0], y_test.shape[2], i * y_test.shape[2] + j + 1)
        plt.plot(X_test[i, 10:, 0], color="gray", alpha=1, label="input sequence")
        plt.plot(y_test[i, 10:, j], color="blue", lw=0.75, label="true target")
        plt.plot(y_pred[i, :, j], color="red", lw=0.75, label="predicted")
        plt.xlabel("time")
        plt.ylabel(f"amplitude of state {j}")
    plt.legend()
plt.show()

# autoRC predictions
# X_test = X_train[:-2, :, :]  # use all but last time step for prediction
# y_test = y_train[:-2, :, :]  # use all but last time step for prediction

###setting time>50 of X_test (last 2 indices ie q1 and q2) =0 so that AutoRC can use it for prediction
X_testn = X_test.copy()
X_testn[:, 500:, 1:] = 0   


feedback_indices = np.array([1,2])  # specify feedback indices here
res_states, y_pred = model_rc.AutoRC_predict(X_testn, fb_scale=1.0, T_run=2200, feedback_indices=feedback_indices)

print(f"shape of input weights: {model_rc.input_layer.weights.shape}")
print(f"shape of feedback weights: {model_rc.feedback_layer.weights.shape}")
print(f"shape of readout weights: {model_rc.readout_layer.weights.shape}")

print(f"input weights:\n{model_rc.input_layer.weights[:5,:]}\n\n"
      f"feedback weights:\n{model_rc.feedback_layer.weights[:5,:]}\n\n"
      f"readout weights:\n{model_rc.readout_layer.weights[:5,:]}")

print(f"shape of predictions on test set: {y_pred.shape}")

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=["mae"])
print(f"Test model loss: {loss_rc}")

# plot predictions vs. ground truth
# r2_scatter(y_true=y_test, y_pred=y_pred)


# plot predictions
plt.figure()
### loop over batch samples in X_test
plt.title("Autonomously running RC predictions")
for i in range(y_test.shape[0]):
    for j in range(y_test.shape[2]):
        plt.subplot(y_test.shape[0], y_test.shape[2], i * y_test.shape[2] + j + 1)
        plt.plot(X_test[i, :, 0], color="gray", alpha=1, label="input sequence")
        plt.plot(y_test[i, :, j], color="blue", lw=0.75, label="true target")
        plt.plot(y_pred[i, :, j], color="red", lw=0.75, label="predicted")
        plt.xlabel("time")
        plt.ylabel(f"amplitude of state {j}")
    plt.legend()

plt.show()


plt.figure()
for i in range(y_pred.shape[0]):
    plt.subplot(y_pred.shape[0], 1, i + 1)  
    plt.plot(y_pred[i, :, 0], y_pred[i, :, 1], color="red", lw=0.75, label="predicted")
plt.show()

"""
k-fold cross-validation of the model: estimate bias and variance of the model
"""
# from pyreco.cross_validation import cross_val

# val, mean, std_dev = cross_val(model_rc, X_train, y_train, n_splits=5, metric=["mse"])
# print(f"Cross-Validation MSE: {val}")
# print(f"Mean MSE: {mean:.3f}")
# print(f"Standard Deviation of MSE: {std_dev:.3f}")



"""
Assigning a custom reservoir matrix to a reservoir computing model.
"""
import numpy as np
from matplotlib import pyplot as plt
from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.utils_data import sequence_to_sequence

"""
Classic RC built on random networks. 

Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal from a sequence of 20 function values
X_train, X_test, y_train, y_test = sequence_to_sequence(
    name="sine_to_cosine",
    n_states=1,
    n_batch=20,
    n_time=50,
)

# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

model = RC()
model.add(InputLayer(input_shape=input_shape))
model.add(
    RandomReservoirLayer(
        nodes=200, density=0.1, activation="tanh", leakage_rate=0.1, fraction_input=0.5
    ),
)
model.add(ReadoutLayer(output_shape, fraction_out=0.99))

# Compile the model
model.compile(discard_transients=10)

# Train the model
model.fit(X_train, y_train)

# Evaluate the model
loss_rc = model.evaluate(X_test, y_test, metrics=["mae"])
print(f"Test model loss: {loss_rc}")

# Make predictions for new data
y_pred = model.predict(X_test)
print(f"shape of predictions on test set: {y_pred.shape}")


"""
Now overwrite the reservoir matrix with a custom one.
"""

weights_old = model.reservoir_layer.weights
spec_rad_old = model.reservoir_layer.spec_rad
density_old = model.reservoir_layer.density

# set the new reservoir weights
weights_new = np.random.randn(200, 200) * 0.1
model.reservoir_layer.set_weights(weights_new)

# we can now also set the spectral radius
model.reservoir_layer.set_spec_rad(0.456789)
spec_rad_new = model.reservoir_layer.spec_rad
density_new = model.reservoir_layer.density

# Train the new model
model.fit(X_train, y_train)

# Evaluate the new model
loss_rc_new = model.evaluate(X_test, y_test, metrics=["mae"])
print(f"Test model loss: {loss_rc}")

# Make predictions for new data
y_pred_new = model.predict(X_test)
print(f"shape of predictions on test set: {y_pred.shape}")

"""
Visual comparison of the predictions with the old and new reservoir weights.
"""

print(f"old reservoir density: {density_old:.3f}, spectral radius: {spec_rad_old:.5f}")
print(f"new reservoir density: {density_new:.3f}, spectral radius: {spec_rad_new:.5f}")

# as we have discarded 10 transients, we can only compare the predictions after that
y_test = y_test[:, 10:, :]
X_test = X_test[:, 10:, :]

plt.figure(figsize=(12, 6))
plt.subplot(1, 2, 1)
plt.scatter(y_test.flatten(), y_pred.flatten(), label="old weights", color='red')
plt.scatter(y_test.flatten(), y_pred_new.flatten(), label="new weights", color="green")
plt.xlabel("ground truth")
plt.ylabel("predictions")
plt.title("Predictions vs. Ground Truth")
plt.legend()

plt.subplot(1, 2, 2)
plt.plot(X_test[0, :, 0], color="gray", alpha=0.3)
plt.plot(X_test[0, :, 0], marker=".", color="black", label="input sequence")
plt.plot(y_test[0, :, 0], marker="o", color="blue", label="true target"
)
plt.plot(y_pred[0, :, 0], marker="x", color="red", 
         label=f"predicted (old weights, density={density_old:.3f}, spectral radius={spec_rad_old:.3f})")
plt.plot(y_pred_new[0, :, 0], marker="v", color="green", 
         label=f"predicted (new weights, density={density_new:.3f}, spectral radius={spec_rad_new:.3f})")
plt.xlabel("time")
plt.ylabel("amplitude")
plt.legend()
plt.title("sequence to scalar prediction")
plt.show()

# figure comparing the network weights
plt.figure(figsize=(12, 6))
plt.subplot(1, 2, 1)
plt.imshow(weights_old, cmap="viridis", aspect="auto")
plt.gca().set_aspect("equal", adjustable="box")
plt.title("Reservoir Weights (Old)")
plt.colorbar()
plt.subplot(1, 2, 2)
plt.imshow(weights_new, cmap="viridis", aspect="auto")
plt.title("Reservoir Weights (New)")
plt.gca().set_aspect("equal", adjustable="box")
plt.colorbar()
plt.tight_layout()
plt.show()






"""
Some testing use case for the CustomModel API of pyreco
"""

from matplotlib import pyplot as plt
from pyreco.custom_models import AutoRC as RC
from pyreco.layers import InputLayer, ReadoutLayer, FeedbackLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.plotting import r2_scatter
from pyreco.utils_data import sequence_to_sequence
from pyreco.optimizers import RidgeSK

"""
Classic RC built on random networks. 

Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal from a sequence of 20 function values
A_train, A_test, B_train, B_test = sequence_to_sequence(
    name="sine_pred",
    n_states=1,
    n_batch=20,
    n_time=501
)

###generate y_train and y_test as previous time step of the input series
X_train = A_train#[:, 1:, :]
X_test = A_test#[:, 1:, :]
### remove last time step from X_train and X_test
y_train = A_train#[:, :-1, :]
y_test = A_test#[:, :-1, :]

print(f"Generated training data shapes: {X_train.shape}, {y_train.shape}")
print(f"Generated test data shapes: {X_test.shape}, {y_test.shape}")

##plot x and y
plt.figure()
for i in range(X_train.shape[0]):
    plt.subplot(X_train.shape[0], 1, i + 1)
    plt.plot(X_train[i, :, 0], color="k", alpha=0.3)
    plt.plot(y_train[i, :, 0], color="b", alpha=0.3)
    plt.xlabel("Time")
    plt.ylabel("Amplitude")
plt.tight_layout()
plt.show()

# print(f"X_train shape: {X_train.shape}, y_train shape: {y_train.shape}")
# print(f"X_test shape: {X_test.shape}, y_test shape: {y_test.shape}")


# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(FeedbackLayer(feedback_shape=output_shape))
model_rc.add(
    RandomReservoirLayer(
        nodes=100, density=0.1, spec_rad=1.0, activation="tanh", leakage_rate=0.2, fraction_input=0.75),
)
model_rc.add(ReadoutLayer(output_shape, fraction_out=0.90))

# Compile the model
optim = RidgeSK(alpha=1.0)
model_rc.AutoRC_compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
    discard_transients=10
)

# Train the model
model_rc.fit(X_train, y_train)

# print(model_rc.input_layer.weights)
# Make predictions for new data
y_pred = model_rc.predict(X_test)

# autoRC predictions
# X_test = X_train[:-2, :, :]  # use all but last time step for prediction
# y_test = y_train[:-2, :, :]  # use all but last time step for prediction
res_states, y_pred = model_rc.AutoRC_predict(X_test, fb_scale=1.0, T_run=3000)

print(f"shape of predictions on test set: {y_pred.shape}")

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=["mae"])
print(f"Test model loss: {loss_rc}")

# plot predictions vs. ground truth
# r2_scatter(y_true=y_test, y_pred=y_pred)


# plot predictions
plt.figure()
### loop over batch samples in X_test
plt.title("Autonomously running RC predictions")
for i in range(X_test.shape[0]):
    plt.subplot(X_test.shape[0], 1, i + 1)
    plt.plot(X_test[i, :, 0], color="gray", alpha=1, label="input sequence")
    plt.plot(y_test[i, :, 0], color="blue", lw=0.75, label="true target")
    plt.plot(y_pred[i, :, 0], color="red", lw=0.75, label="predicted")
    plt.xlabel("time")
    plt.ylabel("amplitude")
    plt.legend()
plt.show()




"""
k-fold cross-validation of the model: estimate bias and variance of the model
"""
# from pyreco.cross_validation import cross_val

# val, mean, std_dev = cross_val(model_rc, X_train, y_train, n_splits=5, metric=["mse"])
# print(f"Cross-Validation MSE: {val}")
# print(f"Mean MSE: {mean:.3f}")
# print(f"Standard Deviation of MSE: {std_dev:.3f}")
